from django.contrib import admin

from .models import *

class TranLineInline(admin.TabularInline):
    model = TranLine

"""
class TransactionAdmin(admin.ModelAdmin):
    list_display=('company', 'id', 'date', 'date_end', 'comment', 'bmo_id', 'source_object')
    list_filter = ('company',)
    ordering = ('-date','id')
    search_fields = ('comment','id',)
    inlines = [
        TranLineInline,
        ]
"""

class TranLineAdmin(admin.ModelAdmin):
    list_display = ('id', 'bmo_id', 'account', 'date', 'amount', 'counterparty',)
    search_fields = ('account__id', 'counterparty__name','id',)
    list_filter = ('account','counterparty')
    search_fields = ('counterparty__name',)
    ordering = ('id',)

class AccountAdmin(admin.ModelAdmin):
    list_display = ('id', 'display_name', 'path', 'ordering', 'role')
    list_filter = ('role',)
    ordering = ('id',)
    save_on_top = True
    save_as = True

class ExternalBalanceAdmin(admin.ModelAdmin):
    list_display= ('company','id','account','date','balance','comment')
    ordering = ('-date','account')
    list_filter = ('company','account',)

class DepreciationPolicyAdmin(admin.ModelAdmin):
    list_display =  ('cap_account', 'depreciation_account', 'depreciation_period',)


class DepartmentAdmin(admin.ModelAdmin):
    ordering = ('id',)
    list_display = ('id', 'name')
    

class CompanyAdmin(admin.ModelAdmin):
    ordering = ('id',)
    list_display = ('id', 'label', 'name', 'cmpy_type')
    
 
class CounterpartyAdmin(admin.ModelAdmin):
    ordering = ('id',)
    list_display = ('id', 'label', 'name',)
    
class ProjectAdmin(admin.ModelAdmin):
    ordering = ('id',)
    list_display = ('id', 'name',)

class EmployeeAdmin(admin.ModelAdmin):
   
    ordering = ('employee_id',)
    list_display = ('employee_id', 'employee_name', 'e_mail', 'department')
    list_filter = ('department',)

class ExternalAccountAdmin(admin.ModelAdmin):
   
    ordering = ('gl_account',)
    list_display = ('gl_account', 'counterparty', 'name', 'label')
    list_filter = ('gl_account','counterparty',)

admin.site.register(Account, AccountAdmin)
#admin.site.register(Transaction, TransactionAdmin)
admin.site.register(TranLine, TranLineAdmin)
admin.site.register(ExternalBalance, ExternalBalanceAdmin)
admin.site.register(DepreciationPolicy, DepreciationPolicyAdmin)
admin.site.register(Department, DepartmentAdmin)
admin.site.register(Company, CompanyAdmin)
admin.site.register(Counterparty, CounterpartyAdmin)
admin.site.register(Project, ProjectAdmin)
admin.site.register(Employee, EmployeeAdmin)
admin.site.register(ExternalAccount, ExternalAccountAdmin)


"""
Partially adapted with permission from ReportLab's DocEngine framework
"""


import requests
import logging
import json
import traceback

from django.views.debug import get_exception_reporter_filter
from django.conf import settings


class DbLogHandler(logging.Handler):
    def emit(self, record):
        from .models import Log
        try:
            request = record.request
            filter = get_exception_reporter_filter(request)
            request_repr = filter.get_request_repr(request)
        except Exception:
            request = None
            request_repr = "Request repr() unavailable."

        if record.exc_info:
            exc_info = record.exc_info
            stack_trace = '\n'.join(traceback.format_exception(*record.exc_info))
        else:
            exc_info = (None, record.getMessage(), None)
            stack_trace = "No stack trace available"

        rec = Log(level=record.levelname,
                  message=record.getMessage(),
                  request=request_repr,
                  traceback=stack_trace
                  )
        if hasattr(record, 'corrId'):
            rec.corrId = record.corrId
        rec.save()


class SlackHandler(logging.Handler):
    def emit(self, record):
        new_data = {"text": '%s: %s' % (record.levelname, record.getMessage())}
        requests.post(settings.SLACK_ENDPOINT_URL, data=json.dumps(new_data))




"""

Based on Andy Robinson's DoubleTalk accounting framework; used
with permission



Cache for general ledger values

This is a singleton object used for all common 'financial queries' -
balances, activity, history.

The name is misleading.  Long ago I needed an in-memory cache and
spent forever worrying about cache invalidation.  With a well indexed
MySQL database, it turned out to be unnecessary; we just do raw
SQL queries for everything we need.


"""
from datetime import date, timedelta
from decimal import Decimal

import pandas as pd
from django.db import connection
from django.conf import settings
from accountifie.common.db import query as qry  #might as well use the standard one
from accountifie.common.api import api_func

DZERO = Decimal('0')

THIS_YEAR_START = date(date.today().year, 1, 1)
THIS_YEAR_END = date(date.today().year + 1, 1, 1) - timedelta(days=1)
INCEPTION = date(2012,12,31)
FOREVER = date(2099,12,31)

# WHAT WE NEED
def get_cache(company_id):
    return GLCache(company_id=company_id)



def normalize_acc(thing):
    if isinstance(thing, str):
        return thing
    elif hasattr(thing, 'id'):  #account
        return thing.id
    else:
        raise ValueError("Expected Account or equivalent text id, but got %s" % repr(thing))

def to_decimal(stuff):
    if stuff is None:
        return DZERO
    else:
        return Decimal(stuff).quantize(DZERO)






class GLCache(object):
    def __init__(self, company_id=None):
        self.company_id = company_id
        self.conn = connection
        self.accts = {}
        self.company_list = api_func('gl', 'company_list', company_id)

    def get_companyID(self):
        return self.company_id

    def query(self, sql, *params):
        return qry(sql, *params)

    ########################################
    # BASIC CHECKS
    def double_entry_check(self):
        "Make sure it adds to zero - return failing transactions"

        sql = """SELECT sum(l.amount)
        FROM gl_tranline l, gl_transaction t
        WHERE t.company_id = ANY(%s)
        AND t.id = l.transaction_id"""
        imbalance = self.query(sql, self.company_list)
        if imbalance:
            #find where

            sql = """SELECT t.id, t.date, t.comment, sum(l.amount)
            FROM gl_transaction t, gl_tranline l
            WHERE t.id = l.transaction_id
            AND t.company_id = ANY(%s)
            GROUP BY t.id
            HAVING sum(l.amount) <> 0
            """
            rows = self.query(sql, self.company_list)
            return rows

    def count(self):
        sql = """SELECT count(t.id) FROM gl_transaction t
              WHERE t.company_id = ANY(%s)"""


        rows = self.query(sql, self.company_list)
        if not rows:
            return 0
        else:
            return rows[0][0]

    def earliest(self):
        sql = "SELECT min(date) from gl_transaction"
        rows = self.query(sql)
        if not rows:
            return date.today()
        else:
            return rows[0][0]

    def latest(self):
        sql = "SELECT max(date) from gl_transaction"
        rows = self.query(sql)
        if not rows:
            return date.today()
        else:
            return rows[0][0]

    def transactions(self):
        # create list of transactions
        # t.comment, t.company_id, t.object_id, t.date
        sql = """SELECT t.id, t.comment, c.name, t.company_id, t.date, t.object_id
        FROM gl_transaction t, django_content_type c
        WHERE t.content_type_id = c.id
        """
        rows = self.query(sql)
        return rows

    def tranlines(self):
        # create list of tranlines
        sql = """SELECT l.id, l.transaction_id, l.amount, a.display_name, a.id, c.name, c.id
        FROM gl_tranline l, gl_account a, gl_counterparty c
        WHERE l.account_id = a.id
        AND c.id = l.counterparty_id
        """
        rows = self.query(sql)
        return rows


    ###################################
    # KILL THIS
    def balance_by_counterparty(self, to_date, account_id):
        "Show the breakdown of the balance. Used for creditors/debtors"
        sql = """
        SELECT l.counterparty_id, sum(l.amount)
        FROM gl_transaction t, gl_tranline l
        WHERE t.id = l.transaction_id
        AND l.account_id = %s
        AND t.date <= %s
        AND t.company_id = ANY(%s)
        GROUP BY l.counterparty_id
        HAVING sum(l.amount) <> 0
        ORDER BY sum(l.amount) DESC
        """
        return self.query(sql, account_id, to_date, self.company_list)

    ########################################
    # NEW STYL GENERIC QUERY

    def get_gl_entries(self, acc_list, from_date=settings.DATE_EARLY, to_date=settings.DATE_LATE):
        sql = """SELECT t.date, t.date_end, t.id, t.comment, l.account_id, l.id, t.company_id, l.counterparty_id, string_agg(c.account_id, ', ') as contra_accounts, string_agg(c.counterparty_id, ', ') as counterparty_ids, l.amount
        FROM gl_tranline l
        INNER JOIN gl_tranline c on c.transaction_id=l.transaction_id
        INNER JOIN gl_transaction t on t.id=l.transaction_id
        WHERE l.account_id = ANY(%s) AND NOT c.account_id = l.account_id
        AND (t.date_end IS NOT NULL OR t.date BETWEEN %s and %s)
        AND (t.date_end IS NULL OR (t.date <= %s AND t.date_end >= %s))
        AND t.company_id = ANY(%s)
        GROUP BY t.id, l.account_id, l.id, t.date, t.comment, l.amount, l.counterparty_id
        ORDER BY t.date;
        """
        rows = self.query(sql,acc_list,from_date, to_date, to_date, from_date, self.company_list)
        if len(rows) == 0:
            return None
        else:
            return pd.DataFrame(rows, columns=['date','date_end','transaction_id','comment','account_id','line_id','company','counterparty','contra_accts','contra_cps','amount']).set_index('line_id')


    def get_gl_entries_trans(self, trans_list, from_date=settings.DATE_EARLY, to_date=settings.DATE_LATE):
        sql = """SELECT t.date, t.date_end, t.id, t.comment, l.account_id, l.id, t.company_id, l.counterparty_id, string_agg(c.account_id, ', ') as contra_accounts, string_agg(c.counterparty_id, ', ') as counterparty_ids, l.amount
        FROM gl_tranline l
        INNER JOIN gl_tranline c on c.transaction_id=l.transaction_id
        INNER JOIN gl_transaction t on t.id=l.transaction_id
        WHERE t.date BETWEEN %s and %s
        AND t.id = ANY(%s)
        GROUP BY t.id, l.account_id, l.id, t.date, t.comment, l.amount, l.counterparty_id
        ORDER BY t.date;
        """
        rows = self.query(sql,from_date, to_date, trans_list)
        if len(rows) == 0:
            return None
        else:
            return pd.DataFrame(rows, columns=['date','date_end','transaction_id','comment','account_id','line_id','company','counterparty','contra_accts','contra_cps','amount']).set_index('line_id')


    def balance(self, acc, date):
        "Balance including entries on given date"
        #whether they passed in a string id or an object, extract the id
        acc = normalize_acc(acc)
        sql = """SELECT sum(l.amount)
        FROM gl_transaction t, gl_tranline l
        WHERE t.id = l.transaction_id
        AND l.account_id = %s
        AND t.date <= %s
        AND t.company_id = ANY(%s)
        """

        rows = self.query(sql, acc, date, self.company_list)

        if not rows:
            return DZERO

        return to_decimal(rows[0][0])




    def get_child_paths(self, path):
        """Return DIRECT child paths.

        Given this, we'd return 'assets.curr.rec'
        even though there is only one rec.clients.  Because I'm lazy
        and it's cleaner for users to understand.
                +---------------------------------+
                | path                            |
                +---------------------------------+
                | assets.curr.ap                  |
                | assets.curr.cashandeq           |
                | assets.curr.fairval.ABSbonds    |
                | assets.curr.fairval.corpbonds   |
                | assets.curr.fairval.equities    |
                | assets.curr.fairval.investments |
                | assets.curr.fairval.loans       |
                | assets.curr.fairval.sovbonds    |
                | assets.curr.other               |
                | assets.curr.rec.clients         |
                | assets.curr.segregated          |
                | assets.curr.stinvest            |
                +---------------------------------+


        """

        from names import DISPLAY_ORDER

        if path.endswith('.'):
            path = path[0:-1]   #we add it ourselves in SQL to remove the current node

        #Can you believe how much escaping we need to get one '%' character through to
        #the database engine?
        sql = "select distinct path, ordering from gl_account where path like '" + path + "%'"
        matches = self.query(sql)


        order_pref = {}
        for (subpath, ordering) in matches:
            order_pref[subpath] = ordering
        order_pref.update(DISPLAY_ORDER)


        nodes = len(path.split("."))
        children = set()
        for row in matches:
            subpath = row[0]
            ordering = row[1]
            parts = subpath.split(".")
            count = len(parts)
            if count == 1 + nodes:
                children.add(subpath)
            elif count >= 1+nodes:
                #e.g. jump from
                #   assets.curr.cashandeq
                #to
                #   assets.curr.cashandeq.checking.jpm
                #without having an actual account for 'checking'
                #then we should infer its existence
                nextbranch = ".".join(parts[0:1+nodes])
                children.add(nextbranch)
        #Take our list of branches and sort in order

        sorter = []
        for subpath in children:
            priority = order_pref.get(subpath, 0)
            sorter.append((priority, subpath))
        sorter.sort()
        return [b for (a,b) in sorter]



from django.contrib import admin
from accountifie.cal.models import Year, Quarter, Month, Week, Day

class YearAdmin(admin.ModelAdmin):
    fields = ('id',)
admin.site.register(Year, YearAdmin)

class QuarterAdmin(admin.ModelAdmin):
    fields = ('id','year',)
    list_filter = ('id','year',)
admin.site.register(Quarter, QuarterAdmin)

class MonthAdmin(admin.ModelAdmin):
    fields = ('id','quarter','year',)
    list_filter = ('quarter','year',)
admin.site.register(Month, MonthAdmin)

class WeekAdmin(admin.ModelAdmin):
    fields = ('id','first_day','last_day','start_month','end_month','year')
    list_filter = ('first_day','last_day','start_month','end_month','year')    
admin.site.register(Week, WeekAdmin)

class DayAdmin(admin.ModelAdmin):
    fields = ('id','month','quarter','year')
list_filter = ('month','quarter','year')    
admin.site.register(Day, DayAdmin)



from colorful.fields import RGBColorField

from django.db import models
from django.core.validators import RegexValidator

ALPHANUMERIC = RegexValidator(r'^[0-9a-zA-Z]*$', 'Only alphanumeric characters are allowed. In particular.. no spaces.')


class Company(models.Model):
    id = models.CharField(max_length=5, primary_key=True)
    label = models.CharField(max_length=20, default='TBD')
    name = models.CharField(max_length=50)
    cmpy_type = models.CharField(max_length=10, choices=[('ALO', 'Standalone'),('CON', 'Consolidation'),])
    color_code = RGBColorField(blank=True)

    subs = models.ManyToManyField('self', blank=True, related_name="+")

    class Meta:
        verbose_name_plural = 'Companies'
        app_label = 'gl'
        db_table = 'gl_company'

    def __str__(self):
        return self.label


class Counterparty(models.Model):
    id = models.CharField(max_length=30, primary_key=True, validators=[ALPHANUMERIC])
    label = models.CharField(max_length=20, default='TBD')
    name = models.CharField(max_length=100)

    class Meta:
        verbose_name_plural = 'Counterparties'
        app_label = 'gl'
        db_table = 'gl_counterparty'

    def __str__(self):
        return self.label

class Project(models.Model):
    id = models.CharField(max_length=12, primary_key=True)
    name = models.CharField(max_length=100)

    class Meta:
        verbose_name_plural = 'Projects'
        app_label = 'gl'
        db_table = 'gl_project'

    def __str__(self):
        return self.id

class Department(models.Model):
    #PK is id, that's OK
    name = models.CharField(max_length=25, unique=True)

    class Meta:
        app_label = 'gl'
        db_table = 'gl_department'

    def __unicode__(self):
        return '%d: %s' % (self.id, self.name)


class Employee(models.Model):
    employee_id = models.IntegerField(primary_key=True)
    ach = models.CharField(max_length=200, null=True)
    usr_disable = models.BooleanField()
    e_mail = models.CharField(max_length=200, null=True)
    role = models.CharField(max_length=200, null=True)
    employee_name = models.CharField(max_length=200, null=True)
    department = models.ForeignKey(Department, on_delete=models.CASCADE)
    p_card = models.CharField(max_length=200, null=True)

    class Meta:
        app_label = 'gl'
        db_table = 'gl_employee'

    def __str__(self):
        return '%s'  % (self.employee_name)

class ExternalAccount(models.Model):
    company = models.ForeignKey('gl.Company', on_delete=models.CASCADE)
    gl_account = models.ForeignKey('gl.Account', on_delete=models.CASCADE)
    counterparty = models.ForeignKey('gl.Counterparty', on_delete=models.CASCADE)
    label = models.CharField(max_length=10)
    name = models.CharField(max_length=100)

    class Meta:
        app_label = 'gl'
        db_table = 'gl_externalaccount'

    def __str__(self):
        return self.label



"""
Adapted with permission from ReportLab's DocEngine framework
"""



from django.conf import settings
from django.conf.urls import include, url
from django.views.generic.base import TemplateView
import django.contrib.auth.views
from accountifie.common.forms import CommonAuthenticationForm, CommonPasswordChangeForm, \
        CommonPasswordResetForm, CommonSetPasswordForm
import accountifie.common.views

urlpatterns = [
    url(r'^login/$', accountifie.common.views.login, {
            'template_name': 'local_login.html',
            'authentication_form': CommonAuthenticationForm,
            #'SSL': True,
            }, name="accounts_login"
            ),
    url(r'^change_password/$', django.contrib.auth.views.password_change, {
            'template_name': 'password_change_form.html',
            'post_change_redirect': '/accounts/change_password/done/',
            'password_change_form': CommonPasswordChangeForm, 
            }, 
            name="change_password"
            ),
    url(r'^change_password/done/$', TemplateView.as_view(template_name='password_change_done.html',),),
    url(r'^logout/$', django.contrib.auth.views.logout, {'template_name': 'common/logout.html', }, 
            name="accounts_logout"
            ),
    url(r'^password_reset/$', django.contrib.auth.views.password_reset, {
            'template_name': 'accounts/password_reset_form.html',
            'email_template_name':'accounts/password_reset_email.html',
            'subject_template_name':'accounts/password_reset_subject.txt',
            'password_reset_form': CommonPasswordResetForm,
            'extra_context': {'REGISTER': 'docengine.register' in settings.INSTALLED_APPS}
            }, 
            name='password_reset'
            ),
    url(r'^password_reset_done/$', django.contrib.auth.views.password_reset_done, 
            {'template_name': 'accounts/password_reset_done.html',},
            name='password_reset_done'
            ),
    url(r'^password_reset_confirm/(?P<uidb64>[0-9A-Za-z]+)-(?P<token>.+)/$', 
            django.contrib.auth.views.password_reset_confirm, 
            {'template_name': 'accounts/password_reset_confirm.html', 
                'set_password_form': CommonSetPasswordForm,},
            name='password_reset_confirm'
            ),
    url(r'^password_reset_complete/$', django.contrib.auth.views.password_reset_complete, 
        {'template_name': 'accounts/password_reset_complete.html',},
        name='password_reset_complete'),
]





from rest_framework import serializers

from accountifie.common.serializers import EagerLoadingMixin

from .models import TranLine


class TranLineSerializer(serializers.ModelSerializer, EagerLoadingMixin):
    _SELECT_RELATED_FIELDS = ['counterparty']
    
    counterparty = serializers.CharField(source='counterparty.name')

    class Meta:
        model = TranLine
        fields = ('id', 'date', 'bmo_id', 'counterparty', 'account_id', 'amount', 'comment')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-12-16 23:04

from __future__ import unicode_literals

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0003_auto_20171127_2240'),
    ]

    operations = [
        migrations.AddField(
            model_name='tranline',
            name='date',
            field=models.DateField(db_index=True, default=datetime.date(2015, 1, 1)),
            preserve_default=False,
        ),
    ]



"""
copied and amended from https://github.com/furious-luke/django-address
"""


from django.db import models
from django.core.exceptions import ValidationError
from django.db.models.fields.related import ForeignObject
try:
    from django.db.models.fields.related_descriptors import ForwardManyToOneDescriptor
except ImportError:
    from django.db.models.fields.related import ReverseSingleRelatedObjectDescriptor as ForwardManyToOneDescriptor
from django.utils.encoding import python_2_unicode_compatible

import logging
logger = logging.getLogger(__name__)

# Python 3 fixes.
import sys
if sys.version > '3':
    long = int
    str = (str, bytes)
    str = str

__all__ = ['Country', 'State', 'Locality', 'Address', 'AddressField']

class InconsistentDictError(Exception):
    pass

def _to_python(value):
    raw = value.get('raw', '')
    country = value.get('country', '')
    country_code = value.get('country_code', '')
    state = value.get('state', '')
    state_code = value.get('state_code', '')
    locality = value.get('locality', '')
    postal_code = value.get('postal_code', '')
    street_number = value.get('street_number', '')
    route = value.get('route', '')
    formatted = value.get('formatted', '')
    latitude = value.get('latitude', None)
    longitude = value.get('longitude', None)

    # If there is no value (empty raw) then return None.
    if not raw:
        return None

    # If we have an inconsistent set of value bail out now.
    if (country or state or locality) and not (country and state and locality):
        raise InconsistentDictError

    # Handle the country.
    try:
        country_obj = Country.objects.get(name=country)
    except Country.DoesNotExist:
        if country:
            if len(country_code) > Country._meta.get_field('code').max_length:
                if country_code != country:
                    raise ValueError('Invalid country code (too long): %s'%country_code)
                country_code = ''
            country_obj = Country.objects.create(name=country, code=country_code)
        else:
            country_obj = None

    # Handle the state.
    try:
        state_obj = State.objects.get(name=state, country=country_obj)
    except State.DoesNotExist:
        if state:
            if len(state_code) > State._meta.get_field('code').max_length:
                if state_code != state:
                    raise ValueError('Invalid state code (too long): %s'%state_code)
                state_code = ''
            state_obj = State.objects.create(name=state, code=state_code, country=country_obj)
        else:
            state_obj = None

    # Handle the locality.
    try:
        locality_obj = Locality.objects.get(name=locality, state=state_obj)
    except Locality.DoesNotExist:
        if locality:
            locality_obj = Locality.objects.create(name=locality, postal_code=postal_code, state=state_obj)
        else:
            locality_obj = None

    # Handle the address.
    try:
        if not (street_number or route or locality):
            address_obj = Address.objects.get(raw=raw)
        else:
            address_obj = Address.objects.get(
                street_number=street_number,
                route=route,
                locality=locality_obj
            )
    except Address.DoesNotExist:
        address_obj = Address(
            street_number=street_number,
            route=route,
            raw=raw,
            locality=locality_obj,
            formatted=formatted,
            latitude=latitude,
            longitude=longitude,
        )

        # If "formatted" is empty try to construct it from other values.
        if not address_obj.formatted:
            address_obj.formatted = str(address_obj)

        # Need to save.
        address_obj.save()

    # Done.
    return address_obj

##
## Convert a dictionary to an address.
##
def to_python(value):

    # Keep `None`s.
    if value is None:
        return None

    # Is it already an address object?
    if isinstance(value, Address):
        return value

    # If we have an integer, assume it is a model primary key. This is mostly for
    # Django being a cunt.
    elif isinstance(value, int):
        return value

    # A string is considered a raw value.
    elif isinstance(value, str):
        obj = Address(raw=value)
        obj.save()
        return obj

    # A dictionary of named address components.
    elif isinstance(value, dict):

        # Attempt a conversion.
        try:
            return _to_python(value)
        except InconsistentDictError:
            return Address.objects.create(raw=value['raw'])

    # Not in any of the formats I recognise.
    raise ValidationError('Invalid address value.')

##
## A country.
##
@python_2_unicode_compatible
class Country(models.Model):
    name = models.CharField(max_length=70, unique=True, blank=True)
    code = models.CharField(max_length=5, blank=True) # not unique as there are duplicates (IT)

    class Meta:
        app_label = 'address'
        db_table = 'address_country'
        verbose_name_plural = 'Countries'
        ordering = ('name',)

    def __str__(self):
        return '%s'%(self.name or self.code)

##
## A state. Google refers to this as `administration_level_1`.
##
@python_2_unicode_compatible
class State(models.Model):
    name = models.CharField(max_length=165, blank=True)
    code = models.CharField(max_length=3, blank=True)
    country = models.ForeignKey(Country, related_name='states')

    class Meta:
        app_label = 'address'
        db_table = 'address_state'
        unique_together = ('name', 'country')
        ordering = ('country', 'name')

    def __str__(self):
        txt = self.to_str()
        country = '%s'%self.country
        if country and txt:
            txt += ', '
        txt += country
        return txt

    def to_str(self):
        return '%s'%(self.name or self.code)

##
## A locality (suburb).
##
@python_2_unicode_compatible
class Locality(models.Model):
    name = models.CharField(max_length=165, blank=True)
    state = models.ForeignKey(State, related_name='localities')

    class Meta:
        app_label = 'address'
        db_table = 'address_locality'
        verbose_name_plural = 'Localities'
        unique_together = ('name', 'state')
        ordering = ('state', 'name')

    def __str__(self):
        txt = '%s'%self.name
        state = self.state.to_str() if self.state else ''
        if txt and state:
            txt += ', '
        txt += state
        if self.postal_code:
            txt += ' %s'%self.postal_code
        cntry = '%s'%(self.state.country if self.state and self.state.country else '')
        if cntry:
            txt += ', %s'%cntry
        return txt

##
## An address. If for any reason we are unable to find a matching
## decomposed address we will store the raw address string in `raw`.
##
@python_2_unicode_compatible
class Address(models.Model):
    street_number = models.CharField(max_length=20, blank=True)
    route = models.CharField(max_length=100, blank=True)
    postal_code = models.CharField(max_length=10, blank=True)
    locality = models.ForeignKey(Locality, related_name='addresses', blank=True, null=True)
    raw = models.CharField(max_length=200)
    formatted = models.CharField(max_length=200, blank=True)
    latitude = models.FloatField(blank=True, null=True)
    longitude = models.FloatField(blank=True, null=True)

    class Meta:
        app_label = 'address'
        db_table = 'address_address'
        verbose_name_plural = 'Addresses'
        ordering = ('locality', 'route', 'street_number')
        # unique_together = ('locality', 'route', 'street_number')

    def __str__(self):
        if self.formatted != '':
            txt = '%s'%self.formatted
        elif self.locality:
            txt = ''
            if self.street_number:
                txt = '%s'%self.street_number
            if self.route:
                if txt:
                    txt += ' %s'%self.route
            locality = '%s'%self.locality
            if txt and locality:
                txt += ', '
            txt += locality
        else:
            txt = '%s'%self.raw
        return txt

    def clean(self):
        if not self.raw:
            raise ValidationError('Addresses may not have a blank `raw` field.')

    def as_dict(self):
        ad = dict(
            street_number=self.street_number,
            route=self.route,
            raw=self.raw,
            formatted=self.formatted,
            latitude=self.latitude if self.latitude else '',
            longitude=self.longitude if self.longitude else '',
        )
        if self.locality:
            ad['locality'] = self.locality.name
            ad['postal_code'] = self.locality.postal_code
            if self.locality.state:
                ad['state'] = self.locality.state.name
                ad['state_code'] = self.locality.state.code
                if self.locality.state.country:
                    ad['country'] = self.locality.state.country.name
                    ad['country_code'] = self.locality.state.country.code
        return ad

class AddressDescriptor(ForwardManyToOneDescriptor):

    def __set__(self, inst, value):
        super(AddressDescriptor, self).__set__(inst, to_python(value))

##
## A field for addresses in other models.
##
class AddressField(models.ForeignKey):
    description = 'An address'

    def __init__(self, **kwargs):
        kwargs['to'] = 'address.Address'
        super(AddressField, self).__init__(**kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False):
        super(ForeignObject, self).contribute_to_class(cls, name, virtual_only=virtual_only)
        setattr(cls, self.name, AddressDescriptor(self))

    # def deconstruct(self):
    #     name, path, args, kwargs = super(AddressField, self).deconstruct()
    #     del kwargs['to']
    #     return name, path, args, kwargs

    def formfield(self, **kwargs):
        from .forms import AddressField as AddressFormField
        defaults = dict(form_class=AddressFormField)
        defaults.update(kwargs)
        return super(AddressField, self).formfield(**defaults)



import accountifie.query.remote_server_status


def backend_lpq(qstring):
    return accountifie.query.remote_server_status.lpq()


def ledger_stats(company_id, qstring):
    return accountifie.query.remote_server_status.gl_stats(company_id)


from decimal import Decimal
from accountifie.query.query_manager import QueryManager
import datetime


def acct_balances(company_id, as_of, filter_closing_entries=False):
    qm = QueryManager()
    if type(as_of) == list:
        dates = as_of
    else:
        dates = [as_of]

    return qm.gl_strategy.account_balances_for_dates(company_id,
                                         None,
                                         dates,
                                         None, None, None,
                                         filter_closing_entries=filter_closing_entries)

def cp_balances(company_id, qstring):
    as_of = qstring.get('date', datetime.datetime.now().date())
    account_id = qstring.get('account')
    qm = QueryManager()
    dates_dict = {'date': as_of}
    return qm.cp_balances(company_id, dates_dict, acct_list=[account_id])[account_id]['date']['closingBalance']

def cp_balance_changes(company_id, qstring):
    dates = qstring.get('date', datetime.datetime.now().date())
    account_id = qstring.get('account')
    qm = QueryManager()
    dates_dict = {'date': dates}

    results = qm.cp_balances(company_id, dates_dict, acct_list=[account_id])[account_id]['date']

    def _get_chg(cp):
        return closing.get(cp, Decimal('0')) - opening.get(cp, Decimal('0'))
    
    closing = dict((l['cp'], Decimal(l['total'])) for l in results['closingBalance'])
    opening = dict((l['cp'], Decimal(l['total'])) for l in results['openingBalance'])
    all_cps = list(set(list(opening.keys()) + list(closing.keys())))
    return [{'total': str(_get_chg(cp)), 'cp': cp} for cp in all_cps]
    